"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from httpx import Response

from .models import InstagramMedia
from .models import MEDIA_FIELDS
from ..utils import HTTPClient
from ..utils.http import _PAYLOAD

if TYPE_CHECKING:
    from .params import InstagramParams



class Instagram:
    """
    Interact with the cloud service API with various methods.

    :param params: Parameters used to instantiate the class.
    """

    __params: 'InstagramParams'
    __client: HTTPClient


    def __init__(
        self,
        params: 'InstagramParams',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__params = params

        client = HTTPClient(
            timeout=params.timeout,
            verify=params.ssl_verify,
            capem=params.ssl_capem)

        self.__client = client


    @property
    def params(
        self,
    ) -> 'InstagramParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def client(
        self,
    ) -> HTTPClient:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client


    def request_block(
        self,
        method: Literal['get'],
        path: str,
        params: Optional[_PAYLOAD] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :returns: Response from upstream request to the server.
        """

        params = dict(params or {})

        server = 'graph.instagram.com'
        client = self.__client

        params['access_token'] = (
            self.__params.token)

        location = (
            f'https://{server}/{path}')

        request = client.request_block

        return request(
            method=method,
            location=location,
            params=params)


    async def request_async(
        self,
        method: Literal['get'],
        path: str,
        params: Optional[_PAYLOAD] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :returns: Response from upstream request to the server.
        """

        params = dict(params or {})

        server = 'graph.instagram.com'
        client = self.__client

        params['access_token'] = (
            self.__params.token)

        location = (
            f'https://{server}/{path}')

        request = client.request_async

        return await request(
            method=method,
            location=location,
            params=params)


    def latest(
        # NOCVR
        self,
    ) -> list[InstagramMedia]:
        """
        Return the posts from the account associated with user.

        :returns: Posts from the account associated with user.
        """

        return self.latest_block()


    def media_block(
        self,
        unique: int | str,
    ) -> InstagramMedia:
        """
        Return the specific content within the social platform.

        :param unique: Unique identifier within social platform.
        :returns: Specific content within the social platform.
        """

        fields = ','.join(MEDIA_FIELDS)


        request = self.request_block

        response = request(
            'get', str(unique),
            {'fields': fields})

        response.raise_for_status()

        fetched = response.json()

        assert isinstance(fetched, dict)


        return InstagramMedia(**fetched)


    async def media_async(
        self,
        unique: int | str,
    ) -> InstagramMedia:
        """
        Return the specific content within the social platform.

        :param unique: Unique identifier within social platform.
        :returns: Specific content within the social platform.
        """

        fields = ','.join(MEDIA_FIELDS)


        request = self.request_async

        response = await request(
            'get', str(unique),
            {'fields': fields})

        response.raise_for_status()

        fetched = response.json()

        assert isinstance(fetched, dict)


        return InstagramMedia(**fetched)


    def latest_block(
        self,
    ) -> list[InstagramMedia]:
        """
        Return the posts from the account associated with user.

        :returns: Posts from the account associated with user.
        """

        fields = ','.join(MEDIA_FIELDS)


        request = self.request_block

        response = request(
            'get', 'me/media',
            {'fields': fields})

        response.raise_for_status()

        fetched = response.json()

        assert isinstance(fetched, dict)


        return [
            InstagramMedia(**x)
            for x in fetched['data']]


    async def latest_async(
        self,
    ) -> list[InstagramMedia]:
        """
        Return the posts from the account associated with user.

        :returns: Posts from the account associated with user.
        """

        fields = ','.join(MEDIA_FIELDS)


        request = self.request_async

        response = await request(
            'get', 'me/media',
            {'fields': fields})

        response.raise_for_status()

        fetched = response.json()

        assert isinstance(fetched, dict)


        return [
            InstagramMedia(**x)
            for x in fetched['data']]



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path

from .helpers import ByteStreamAsync
from .helpers import ByteStreamBlock



SAMPLES = (
    Path(__file__).parent
    / 'samples')



__all__ = [
    'ByteStreamBlock',
    'ByteStreamAsync']



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from encommon.types import DictStrAny
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils import read_sample
from encommon.utils import read_text
from encommon.utils.sample import ENPYRWS

from httpx import Response

from pytest import fixture
from pytest import mark

from respx import MockRouter

from . import ByteStreamAsync
from . import ByteStreamBlock
from . import SAMPLES
from ..bridge import Bridge
from ..params import BridgeParams



@fixture
def bridge() -> Bridge:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    params = BridgeParams(
        server='192.168.1.10',
        token='mocked')

    return Bridge(params)



def test_Bridge(
    bridge: Bridge,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param bridge: Class instance for connecting to service.
    """


    attrs = lattrs(bridge)

    assert attrs == [
        '_Bridge__params',
        '_Bridge__client']


    assert inrepr(
        'bridge.Bridge object',
        bridge)

    assert isinstance(
        hash(bridge), int)

    assert instr(
        'bridge.Bridge object',
        bridge)


    assert bridge.params

    assert bridge.client



def test_Bridge_request(
    bridge: Bridge,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param bridge: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    source = read_text(
        f'{SAMPLES}/source'
        '/resource.json')

    source = read_sample(
        sample=source)


    (respx_mock
     .get(
         'https://192.168.1.10'
         '/clip/v2/resource')
     .mock(Response(
         status_code=200,
         content=source)))


    request = bridge.request

    response = request(
        method='get',
        path='resource')

    response.raise_for_status()

    fetched = response.json()


    sample_path = (
        f'{SAMPLES}/dumped'
        '/resource.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=fetched)

    expect = prep_sample(
        content=fetched)

    assert expect == sample



def test_Bridge_events_block(
    bridge: Bridge,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param bridge: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    source = read_text(
        f'{SAMPLES}/source'
        '/events.json')

    source = read_sample(
        sample=source)

    streamer = (
        ByteStreamBlock(source))

    (respx_mock
     .get(
         'https://192.168.1.10'
         '/eventstream/clip/v2')
     .mock(Response(
         status_code=200,
         stream=streamer)))


    request = bridge.events_block

    events = list(request())


    sample_path = (
        f'{SAMPLES}/dumped'
        '/events.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=events)

    expect = prep_sample(
        content=events)

    assert expect == sample



@mark.asyncio
async def test_Bridge_events_async(
    bridge: Bridge,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param bridge: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    source = read_text(
        f'{SAMPLES}/source'
        '/events.json')

    source = read_sample(
        sample=source)

    streamer = (
        ByteStreamAsync(source))

    (respx_mock
     .get(
         'https://192.168.1.10'
         '/eventstream/clip/v2')
     .mock(Response(
         status_code=200,
         stream=streamer)))


    request = bridge.events_async

    _events = request()

    events: list[DictStrAny] = []

    async for event in _events:
        events.append(event)


    sample_path = (
        f'{SAMPLES}/dumped'
        '/events.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=events)

    expect = prep_sample(
        content=events)

    assert expect == sample



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Thread
from time import sleep as block_sleep

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from pytest import raises

from .helpers import EVENTS
from .helpers import SVENTS
from ..client import Client
from ..models import ClientEvent
from ..params import ClientParams
from ...fixtures import IRCClientSocket



def test_ClientEvent() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    params = ClientParams(
        server='mocked',
        port=6667,
        nickname='ircbot',
        username='ircbot',
        realname='ircbot',
        ssl_enable=False)

    client = Client(params)


    _event = (
        ':server PING'
        ' :123456789')

    event = ClientEvent(
        client, _event)


    attrs = lattrs(event)

    assert attrs == [
        'prefix',
        'command',
        'params',
        'original',
        'kind',
        'isme',
        'author',
        'recipient',
        'message']


    assert inrepr(
        'ClientEvent(prefix',
        event)

    with raises(TypeError):
        hash(event)

    assert instr(
        "prefix='server'",
        event)


    assert event.original

    assert event.kind == 'event'

    assert not event.isme

    assert not event.author

    assert not event.recipient

    assert not event.message



def test_ClientEvent_cover(  # noqa: CFQ001
    client_ircsock: IRCClientSocket,
) -> None:
    """
    Perform various tests associated with relevant routines.

    .. note::
       Duplicative test routine as below but with EVENTS.

    :param client_ircsock: Object to mock client connection.
    """

    params = ClientParams(
        server='mocked',
        port=6667,
        nickname='ircbot',
        username='ircbot',
        realname='ircbot',
        ssl_enable=False)

    client = Client(params)


    def _operate() -> None:

        client_ircsock(EVENTS)

        _raises = ConnectionError

        with raises(_raises):
            client.operate()


    thread = Thread(
        target=_operate)

    thread.start()


    mqueue = client.mqueue


    item = mqueue.get()

    assert item.prefix == 'mocked'
    assert item.command == '001'
    assert item.params == (
        'ircbot :Welcome to network')

    assert item.kind == 'event'
    assert not item.isme
    assert not item.author
    assert not item.recipient
    assert not item.message

    assert not client.canceled
    assert client.connected
    assert client.nickname == 'ircbot'


    item = mqueue.get()

    assert item.prefix == 'mocked'
    assert item.command == '376'
    assert item.params == (
        'ircbot :End of /MOTD command.')

    assert item.kind == 'event'
    assert not item.isme
    assert not item.author
    assert not item.recipient
    assert not item.message


    item = mqueue.get()

    assert item.prefix == 'mocked'
    assert item.command == '376'
    assert item.params == (
        'ircbot :End of /MOTD command.')

    assert item.kind == 'event'
    assert not item.isme
    assert not item.author
    assert not item.recipient
    assert not item.message


    item = mqueue.get()

    assert item.prefix == 'mocked'
    assert item.command == '376'
    assert item.params == (
        'ircbot :End of /MOTD command.')

    assert item.kind == 'event'
    assert not item.isme
    assert not item.author
    assert not item.recipient
    assert not item.message


    item = mqueue.get()

    assert item.prefix == (
        'nick!user@host')
    assert item.command == 'PRIVMSG'
    assert item.params == (
        'ircbot :Hello ircbot')

    assert item.kind == 'privmsg'
    assert not item.isme
    assert item.author == 'nick'
    assert item.recipient == 'ircbot'
    assert item.message == (
        'Hello ircbot')


    item = mqueue.get()

    assert item.prefix == (
        'nick!user@host')
    assert item.command == 'PRIVMSG'
    assert item.params == (
        '# :Hello world')

    assert item.kind == 'chanmsg'
    assert not item.isme
    assert item.author == 'nick'
    assert item.recipient == '#'
    assert item.message == (
        'Hello world')


    item = mqueue.get()

    assert item.prefix == (
        'nick!user@host')
    assert item.command == 'PRIVMSG'
    assert item.params == (
        '#funchat :Hello world')

    assert item.kind == 'chanmsg'
    assert not item.isme
    assert item.author == 'nick'
    assert item.recipient == '#funchat'
    assert item.message == (
        'Hello world')


    item = mqueue.get()

    assert item.prefix == (
        'ircbot!user@host')
    assert item.command == 'NICK'
    assert item.params == ':botirc'

    assert item.kind == 'event'
    assert not item.isme
    assert not item.author
    assert not item.recipient
    assert not item.message

    assert client.nickname == 'botirc'


    item = mqueue.get()

    assert item.prefix == (
        'botirc!user@host')
    assert item.command == 'PRIVMSG'
    assert item.params == (
        '# :Hello nick')

    assert item.kind == 'chanmsg'
    assert item.isme
    assert item.author == 'botirc'
    assert item.recipient == '#'
    assert item.message == (
        'Hello nick')


    item = mqueue.get()

    assert not item.prefix
    assert item.command == 'ERROR'
    assert item.params == (
        ':Closing Link: botirc'
        '[mocked] (Quit: botirc)')

    assert item.kind == 'event'
    assert not item.isme
    assert not item.author
    assert not item.recipient
    assert not item.message

    block_sleep(1)

    assert not client.canceled
    assert not client.connected
    assert client.nickname == 'botirc'


    thread.join(10)


    assert mqueue.empty()



def test_ClientEvent_service(
    client_ircsock: IRCClientSocket,
) -> None:
    """
    Perform various tests associated with relevant routines.

    .. note::
       Duplicative test routine as above but with SVENTS.

    :param client_ircsock: Object to mock client connection.
    """

    sname = 'jupiter.enasis.net'
    about = 'Network Services'

    params = ClientParams(
        server='mocked',
        port=6900,
        operate='service',
        servername=sname,
        password='password',
        realname=about)

    client = Client(params)


    def _operate() -> None:

        client_ircsock(SVENTS)

        _raises = ConnectionError

        with raises(_raises):
            client.operate()


    thread = Thread(
        target=_operate)

    thread.start()


    mqueue = client.mqueue


    for count in range(20):

        item = mqueue.get()

        assert not item.prefix
        assert not item.command
        assert not item.params

        assert item.kind == 'event'

        assert (
            not item.isme
            if count != 13
            else item.isme)

        assert not item.author
        assert not item.recipient
        assert not item.message


    assert client.nickname == '42X'


    thread.join(10)


    assert mqueue.empty()



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any



def dumlog(
    **kwargs: Any,
) -> None:
    """
    Capture and drop messages for would be provided logger.

    :param kwargs: Keyword arguments which would be passed.
    """



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .models import YouTubeResult
from .models import YouTubeVideo
from .params import YouTubeParams
from .youtube import YouTube



__all__ = [
    'YouTube',
    'YouTubeParams',
    'YouTubeResult',
    'YouTubeVideo']



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .models import RedditListing
from .params import RedditParams
from .reddit import Reddit



__all__ = [
    'Reddit',
    'RedditParams',
    'RedditListing']



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import dumps
from json import loads

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils import read_text
from encommon.utils.sample import ENPYRWS

from httpx import Response

from pytest import fixture
from pytest import mark

from respx import MockRouter

from . import SAMPLES
from ..instagram import Instagram
from ..params import InstagramParams



@fixture
def social() -> Instagram:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    params = InstagramParams(
        token='mocked')

    return Instagram(params)



def test_Instagram(
    social: Instagram,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    """


    attrs = lattrs(social)

    assert attrs == [
        '_Instagram__params',
        '_Instagram__client']


    assert inrepr(
        'instagram.Instagram object',
        social)

    assert isinstance(
        hash(social), int)

    assert instr(
        'instagram.Instagram object',
        social)


    assert social.params

    assert social.client



def test_Instagram_block(
    social: Instagram,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _latest = read_text(
        SAMPLES / 'source.json')

    _media = dumps(loads(
        _latest)['data'][0])

    location = (
        'https://graph.instagram.com')


    (respx_mock
     .get(f'{location}/me/media')
     .mock(Response(
         status_code=200,
         content=_latest)))

    (respx_mock
     .get(f'{location}/mocked')
     .mock(Response(
         status_code=200,
         content=_media)))


    latest = (
        social.latest_block())

    media = (
        social.media_block('mocked'))


    sample_path = (
        SAMPLES / 'latest.json')

    sample = load_sample(
        sample_path,
        [x.endumped
         for x in latest],
        update=ENPYRWS)

    expect = prep_sample([
        x.endumped
        for x in latest])

    assert expect == sample


    sample_path = (
        SAMPLES / 'media.json')

    sample = load_sample(
        sample_path,
        media.endumped,
        update=ENPYRWS)

    expect = prep_sample(
        media.endumped)

    assert expect == sample



@mark.asyncio
async def test_Instagram_async(
    social: Instagram,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _latest = read_text(
        SAMPLES / 'source.json')

    _media = dumps(loads(
        _latest)['data'][0])

    location = (
        'https://graph.instagram.com')


    (respx_mock
     .get(f'{location}/me/media')
     .mock(Response(
         status_code=200,
         content=_latest)))

    (respx_mock
     .get(f'{location}/mocked')
     .mock(Response(
         status_code=200,
         content=_media)))


    latest = await (
        social.latest_async())

    media = await (
        social.media_async('mocked'))


    sample_path = (
        SAMPLES / 'latest.json')

    sample = load_sample(
        sample_path,
        [x.endumped
         for x in latest],
        update=ENPYRWS)

    expect = prep_sample([
        x.endumped
        for x in latest])

    assert expect == sample


    sample_path = (
        SAMPLES / 'media.json')

    sample = load_sample(
        sample_path,
        media.endumped,
        update=ENPYRWS)

    expect = prep_sample(
        media.endumped)

    assert expect == sample



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import loads

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import read_text

from pytest import raises

from . import SAMPLES
from ..models import RedditListing



def test_RedditListing() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    content = read_text(
        SAMPLES / 'source.json')

    loaded = loads(content)

    assert isinstance(
        loaded, dict)

    source = (
        loaded['data']
        ['children']
        [0]['data'])


    item = (
        RedditListing(**source))


    attrs = lattrs(item)

    assert attrs == [
        'name',
        'id',
        'created',
        'title',
        'selftext',
        'author',
        'url',
        'permalink',
        'thumbnail',
        'url_dest',
        'domain',
        'medias',
        'pinned',
        'edited',
        'stickied',
        'archived',
        'vote_downs',
        'vote_ups',
        'score']


    assert inrepr(
        'RedditListing(name',
        item)

    with raises(TypeError):
        hash(item)

    assert instr(
        "name='t3_1c2wc19'",
        item)


    assert item.name == 't3_1c2wc19'

    assert item.id == '1c2wc19'

    assert item.created == 1712993526

    assert item.title[:7] == 'Welcome'

    assert item.selftext
    assert item.selftext[:5] == 'Hello'

    assert item.author == 'enasisnetwork'

    assert item.url[:5] == 'https'

    assert item.permalink[:16] == '/r/enasisnetwork'

    assert item.thumbnail == 'self'

    assert item.url_dest is None

    assert item.domain == 'self.enasisnetwork'

    assert item.medias is None

    assert item.pinned is False

    assert item.edited is False

    assert item.stickied is False

    assert item.archived is False

    assert item.vote_downs == 0

    assert item.vote_ups == 1

    assert item.score == 1



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import TYPE_CHECKING

from httpx import Response

from ..utils import HTTPClient
from ..utils.http import _METHODS
from ..utils.http import _PAYLOAD

if TYPE_CHECKING:
    from .params import RouterParams



class Router:
    """
    Interact with the local service API with various methods.

    :param params: Parameters used to instantiate the class.
    """

    __params: 'RouterParams'
    __client: HTTPClient


    def __init__(
        self,
        params: 'RouterParams',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__params = params

        client = HTTPClient(
            timeout=params.timeout,
            verify=params.ssl_verify,
            capem=params.ssl_capem)

        self.__client = client


    @property
    def params(
        self,
    ) -> 'RouterParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def client(
        self,
    ) -> HTTPClient:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client


    def request_cookie(
        self,
    ) -> Response:
        """
        Establish new session obtaining cookie for authorization.

        :returns: Response from upstream request to the server.
        """

        _params = self.__params
        username = _params.username
        password = _params.password

        payload = {
            'username': username,
            'password': password}

        response = self.request(
            method='post',
            path='api/auth/login',
            json=payload)

        response.raise_for_status()

        return response


    def request(
        self,
        method: _METHODS,
        path: str,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
        *,
        timeout: Optional[int] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        :param timeout: Timeout waiting for the server response.
            This will override the default client instantiated.
        :returns: Response from upstream request to the server.
        """

        params = dict(params or {})
        json = dict(json or {})

        _params = self.__params
        server = _params.server

        client = self.__client

        location = (
            f'https://{server}/{path}')

        request = client.request_block

        return request(
            method=method,
            location=location,
            params=params,
            json=json)


    def reqroxy(
        self,
        method: _METHODS,
        path: str,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
        *,
        timeout: Optional[int] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        .. note::
           Most endpoints on Ubiquiti routers are prefixed with
           not only the site but also `/proxy/network/api`. The
           method will handle concatenating the values for you.
           It will also handle authenticating if not already.

        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        :param timeout: Timeout waiting for the server response.
            This will override the default client instantiated.
        :returns: Response from upstream request to the server.
        """

        _params = self.__params
        site = _params.site

        path = (
            'proxy/network/api'
            f'/s/{site}/{path}')


        def _request() -> Response:

            return self.request(
                method, path,
                params, json)


        response = _request()

        if response.status_code == 401:

            self.request_cookie()

            response = _request()

        response.raise_for_status()


        return response



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import TYPE_CHECKING

from httpx import Response

from ..utils import HTTPClient
from ..utils.http import _METHODS
from ..utils.http import _PAYLOAD

if TYPE_CHECKING:
    from .params import BridgeParams



class Bridge:
    """
    Interact with the local service API with various methods.

    :param params: Parameters used to instantiate the class.
    """

    __params: 'BridgeParams'
    __client: HTTPClient


    def __init__(
        self,
        params: 'BridgeParams',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__params = params

        client = HTTPClient(
            timeout=params.timeout,
            verify=params.ssl_verify,
            capem=params.ssl_capem)

        self.__client = client


    @property
    def params(
        self,
    ) -> 'BridgeParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def client(
        self,
    ) -> HTTPClient:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client


    def request(
        self,
        method: _METHODS,
        path: str,
        params: Optional[_PAYLOAD] = None,
        *,
        timeout: Optional[int] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param timeout: Timeout waiting for the server response.
            This will override the default client instantiated.
        :returns: Response from upstream request to the server.
        """

        params = dict(params or {})

        _params = self.__params
        server = _params.server
        appid = _params.appid
        token = _params.token

        client = self.__client

        tokey = 'Authorization'
        token = f'Bearer {token}'

        headers = {tokey: token}

        location = (
            f'https://{server}/apps'
            f'/api/{appid}/{path}')

        request = client.request_block

        return request(
            method=method,
            location=location,
            params=params,
            headers=headers)



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from unittest.mock import Mock

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from pytest import fixture
from pytest import raises

from ..client import Client
from ..params import ClientParams
from ...fixtures import MTMClientSocket



@fixture
def client() -> Client:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    params = ClientParams(
        server='mocked',
        token='mocked',
        teamid='mocked')

    return Client(params)



def test_Client(
    client: Client,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param client: Class instance for connecting to service.
    """


    attrs = lattrs(client)

    assert attrs == [
        '_Client__params',
        '_Client__logger',
        '_Client__client',
        '_Client__socket',
        '_Client__conned',
        '_Client__exited',
        '_Client__mynick',
        '_Client__lsnick',
        '_Client__mqueue',
        '_Client__cancel']


    assert inrepr(
        'client.Client object',
        client)

    assert isinstance(
        hash(client), int)

    assert instr(
        'client.Client object',
        client)


    assert client.params

    assert not client.connected

    assert not client.nickname

    assert client.mqueue.qsize() == 0

    assert not client.canceled



def test_Client_cover(
    client: Client,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param client: Class instance for connecting to service.
    """

    client.stop()

    client._Client__conned.set()  # type: ignore
    client._Client__socket = Mock()  # type: ignore

    client.stop()



def test_Client_connect(
    client: Client,
    client_mtmsock: MTMClientSocket,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param client: Class instance for connecting to service.
    :param client_mtmsock: Object to mock client connection.
    """

    client_mtmsock()


    with raises(ConnectionError):
        client.operate()

    assert not client.canceled
    assert not client.connected

    mqueue = client.mqueue

    assert mqueue.qsize() == 6



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Optional

from encommon.types import BaseModel

from pydantic import Field



MEDIA_FIELDS = [
    'caption',
    'id',
    'is_shared_to_feed',
    'media_type',
    'media_url',
    'permalink',
    'thumbnail_url',
    'timestamp',
    'username']

MEDIA_RENAME = {
    'is_shared_to_feed': 'shared',
    'media_type': 'type',
    'media_url': 'location',
    'thumbnail_url': 'thumbnail'}



class InstagramMedia(BaseModel, extra='allow'):
    """
    Contains information returned from the upstream response.
    """

    caption: Annotated[
        Optional[str],
        Field(None,
              description='Value from the server response')]

    id: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    shared: Annotated[
        Optional[bool],
        Field(None,
              description='Value from the server response')]

    type: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    location: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    permalink: Annotated[
        Optional[str],
        Field(None,
              description='Value from the server response')]

    thumbnail: Annotated[
        Optional[str],
        Field(None,
              description='Value from the server response')]

    timestamp: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    username: Annotated[
        str,
        Field(...,
              description='Value from the server response')]


    def __init__(
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        items = MEDIA_RENAME.items()

        for old, new in items:

            if old not in data:
                continue

            data[new] = data[old]

            del data[old]

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ssl import SSLContext
from ssl import SSLSocket
from time import sleep as block_sleep
from typing import Iterator
from typing import Optional
from typing import Protocol
from typing import overload
from unittest.mock import MagicMock
from unittest.mock import Mock

from pytest import fixture

from pytest_mock import MockerFixture



_EVENTS = Optional[list[str]]

_SOCKET = tuple[
    SSLContext,
    MagicMock]

_TUBLYTES = tuple[bytes, ...]



EVENTS: list[str] = [

    (':mocked 376 ircbot '
     ':End of /MOTD command.'),

    (':nick!user@host PRIVMSG'
     ' ircbot :Hello ircbot'),

    (':nick!user@host PRIVMSG'
     ' # :Hello world'),

    (':nick!user@host PRIVMSG'
     ' #funchat :Hello world'),

    (':ircbot!user@host'
     ' NICK :botirc'),

    (':botirc!user@host PRIVMSG'
     ' # :Hello nick'),

    ('ERROR :Closing Link: botirc'
     '[mocked] (Quit: botirc)')]



SVENTS: list[str] = [

    ('SERVER jupiter.invalid 1'
     ' :U6100-Fhn6OoErmM-0AD IRC Network'),

    (':0AD SID neptune.invalid'
     ' 2 0PS :IRC Network'),

    ('@s2s-md/tls_cipher='
     'TLSv1.3-TLS_CHACHA20_POLY1305_SHA256;'
     's2s-md/creationtime=1727476461'
     ' :0PS UID ircbot 0'
     ' 1727476461 ircbot'
     ' localhost 0PS3FGG02'
     ' 0 +iwxz * irc-5DE79E97'
     ' fwAAAQ== :ircbot'),

    (':0AD MD client 0PS3FGG02'
     ' creationtime :1727476461'),

    (':0AD MD client'
     ' 0PS3FGG02 tls_cipher '
     ':TLSv1.3-TLS_CHACHA20_POLY1305_SHA256'),

    ('@s2s-md/tls_cipher='
     'TLSv1.3-TLS_CHACHA20_POLY1305_SHA256;'
     's2s-md/creationtime=1727476468;'
     's2s-md/operlogin=robert;'
     's2s-md/operclass=netadmin'
     ' :0AD UID robert 0'
     ' 1727508036 robert'
     ' localhost 0AD1QCV03'
     ' 0 +iowxz * irc-5DE79E97'
     ' AAAAAAAAAAAAAAAAAAAAAQ== :robert'),

    (':0AD MD client 0AD1QCV03'
     ' operclass :netadmin'),

    (':0AD MD client 0AD1QCV03'
     ' operlogin :robert'),

    (':0AD MD client 0AD1QCV03'
     ' creationtime :1727476468'),

    (':0AD MD client'
     ' 0AD1QCV03 tls_cipher '
     ':TLSv1.3-TLS_CHACHA20_POLY1305_SHA256'),

    (':42X UID ChatServ 0'
     ' 1727569796 service'
     ' services.invalid 42X000009'
     ' 0 +Sio services.invalid'
     ' services.invalid * '
     ':Invitable LLM chatting'),

    ('@msgid=Z48DNGsGfMhR8UwXhAljGz;'
     'time=2024-09-29T00:29:56.463Z'
     ' :0AD SJOIN 1727476468'
     ' #opers +nt :@0AD1QCV03'),

    ('NETINFO 11 1727569796 6100'
     ' SHA256:b94e96b5d7fbba92fc0a36c2'
     '667d4d1581a6281ac1440a741d5556f3'
     ' 0 0 0 :EnasisNET'),

    (':0AD1QCV03 SAJOIN'
     ' 0PS3FGG02 #funchat'),

    ('@time=2024-09-29T00:53:42.190Z;'
     'msgid=bdVQ13oQfRcdUnxX2hbb4C;'
     'unrealircd.org/issued-by='
     'OPER:robert@jupiter.invalid:robert'
     ' :0PS SJOIN 1727571222'
     ' #funchat :0PS3FGG02'),

    ('@unrealircd.org/issued-by='
     'OPER:robert@jupiter.invalid:robert;'
     'msgid=bdVQ13oQfRcdUnxX2hbb4C'
     '-6D9mg3IKCGFgtCeraInDIA;'
     'time=2024-09-29T00:53:42.190Z'
     ' :0PS MODE #funchat'
     ' +nt  1727571222'),

    ('@msgid=mjvxur6dNstbXFajGoPjrU;'
     'time=2024-09-29T00:53:45.577Z'
     ' :0AD SJOIN 1727571222'
     ' #funchat :0AD1QCV03')]



RVENTS: list[str] = [

    (':mocked 001 ircbot'
     ' :Welcome to network'),

    (':mocked 376 ircbot '
     ':End of /MOTD command.'),

    'PING :123456789',

    (':mocked 376 ircbot '
     ':End of /MOTD command.'),

    'PING :123456789']



class IRCClientSocket(Protocol):
    """
    Typing protocol which the developer does not understand.
    """

    @overload
    def __call__(
        self,
        rvents: _EVENTS,
    ) -> _SOCKET:
        ...  # NOCVR

    @overload
    def __call__(
        self,
    ) -> _SOCKET:
        ...  # NOCVR

    def __call__(
        self,
        rvents: _EVENTS = None,
    ) -> _SOCKET:
        """
        Construct the instance for use in the downstream tests.

        :param rvents: Raw events for playback from the server.
        """
        ...  # NOCVR



@fixture
def client_ircsock(  # noqa: CFQ004
    mocker: MockerFixture,
) -> IRCClientSocket:
    """
    Construct the instance for use in the downstream tests.

    :param mocker: Object for mocking the Python routines.
    :returns: Newly constructed instance of related class.
    """

    mckctx = mocker.patch(
        ('enconnect.irc'
         '.client.default'),
        autospec=True)

    mckmod = mocker.patch(
        ('enconnect.irc'
         '.client.socket'),
        autospec=True)

    secket = (
        mckctx.return_value)

    secmod = (
        secket.wrap_socket)


    def _split(
        event: str,
    ) -> _TUBLYTES:

        event += '\r\n'

        split = [
            x.encode('utf-8')
            for x in event]

        return tuple(split)


    def _encode(
        resps: list[str],
    ) -> list[_TUBLYTES]:

        items = [
            _split(x)
            for x in resps]

        return items


    def _delayed(
        events: list[_TUBLYTES],
    ) -> Iterator[bytes]:

        while True:

            for event in events:

                block_sleep(0.1)

                yield from event

            block_sleep(0.1)

            yield from [b'']


    def _factory(
        rvents: list[str],
    ) -> MagicMock:

        effect = _delayed(
            _encode(rvents))

        socket = MagicMock(
            SSLSocket)

        socket.send = Mock()

        socket.recv = Mock(
            side_effect=effect)

        socket.close = Mock()

        return socket


    def _fixture(
        rvents: _EVENTS = None,
    ) -> _SOCKET:

        rvents = rvents or []

        socket = _factory(
            RVENTS + rvents)

        secmod.return_value = socket
        mckmod.return_value = socket

        return (secket, socket)


    return _fixture



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Optional

from encommon.types import BaseModel
from encommon.types import getate
from encommon.types import sort_dict

from pydantic import Field



LISTING_VALUE = {
    'created_utc': 'created',
    'ups': 'vote_ups',
    'downs': 'vote_downs',
    'url_overridden_by_dest': 'url_dest',
    'media_metadata': 'medias'}



class RedditListing(BaseModel, extra='ignore'):
    """
    Contains information returned from the upstream response.
    """

    name: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    id: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    created: Annotated[
        int,
        Field(...,
              description='Value from the server response')]

    title: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    selftext: Annotated[
        Optional[str],
        Field(None,
              description='Value from the server response')]

    author: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    url: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    permalink: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    thumbnail: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    url_dest: Annotated[
        Optional[str],
        Field(None,
              description='Value from the server response')]

    domain: Annotated[
        str,
        Field(...,
              description='Value from the server response')]

    medias: Annotated[
        Optional[list[str]],
        Field(None,
              description='Value from the server response')]

    pinned: Annotated[
        bool,
        Field(...,
              description='Value from the server response')]

    edited: Annotated[
        bool | float,
        Field(...,
              description='Value from the server response')]

    stickied: Annotated[
        bool,
        Field(...,
              description='Value from the server response')]

    archived: Annotated[
        bool,
        Field(...,
              description='Value from the server response')]

    vote_downs: Annotated[
        int,
        Field(...,
              description='Value from the server response')]

    vote_ups: Annotated[
        int,
        Field(...,
              description='Value from the server response')]

    score: Annotated[
        int,
        Field(...,
              description='Value from the server response')]


    def __init__(
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        items = LISTING_VALUE.items()

        for old, new in items:

            value = data.get(old)

            if value is None:
                continue

            data[new] = value


        data = {
            k: v for k, v in
            data.items()
            if v not in ['', None]}


        if data.get('medias'):

            images: list[str] = []

            medias = data['medias']


            gitems = getate(
                data,
                'gallery_data/items')


            if gitems is not None:

                for item in gitems:

                    _unique = item['id']
                    _media = item['media_id']

                    if _media not in medias:
                        continue  # NOCVR

                    media = medias[_media]

                    medias[_unique] = media

                    del medias[_media]


            mitems = (
                sort_dict(medias)
                .items())


            for _, media in mitems:

                assert isinstance(
                    media, dict)

                image = getate(
                    media, 's/u')

                assert isinstance(
                    image, str)

                images.append(image)


            data['medias'] = images




        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .instagram import Instagram
from .models import InstagramMedia
from .params import InstagramParams



__all__ = [
    'Instagram',
    'InstagramParams',
    'InstagramMedia']



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from re import compile
from re import match as re_match
from typing import Annotated
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import BaseModel
from encommon.types import DictStrAny
from encommon.types import NCTrue

from pydantic import Field

if TYPE_CHECKING:
    from .client import Client



EVENT = compile(
    r'^(?::(?P<prefix>[^\s]+)\s)?'
    r'(?P<command>[A-Z0-9]{3,})\s'
    r'(?P<params>.+)$')

KINDS = Literal[
    'event',
    'chanmsg',
    'privmsg']

MESSAGE = [
    'chanmsg',
    'privmsg']



class ClientEvent(BaseModel, extra='ignore'):
    """
    Contains information returned from the upstream server.
    """

    prefix: Annotated[
        Optional[str],
        Field(None,
              description='Prefix or origin information',
              min_length=1)]

    command: Annotated[
        Optional[str],
        Field(None,
              description='Code or command for the event',
              min_length=1)]

    params: Annotated[
        Optional[str],
        Field(None,
              description='Event or command parameters')]

    original: Annotated[
        str,
        Field(...,
              description='Original received from server',
              min_length=1)]

    kind: Annotated[
        KINDS,
        Field('event',
              description='Dynamic field parsed from event')]

    isme: Annotated[
        bool,
        Field(False,
              description='Indicates message is from client')]

    author: Annotated[
        Optional[str],
        Field(None,
              description='Dynamic field parsed from event',
              min_length=1)]

    recipient: Annotated[
        Optional[str],
        Field(None,
              description='Dynamic field parsed from event',
              min_length=1)]

    message: Annotated[
        Optional[str],
        Field(None,
              description='Dynamic field parsed from event',
              min_length=1)]


    def __init__(
        self,
        /,
        client: 'Client',
        event: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        data: DictStrAny = {
            'original': event}

        operate = (
            client.params
            .operate)


        match = re_match(
            EVENT, event)

        if (operate == 'normal'
                and match is not None):

            prefix = (
                match
                .group('prefix'))

            if prefix is not None:
                data['prefix'] = (
                    prefix.strip())

            command = (
                match
                .group('command'))

            if command is not None:
                data['command'] = (
                    command.strip())

            params = (
                match
                .group('params'))

            if params is not None:
                data['params'] = (
                    params.strip())


        super().__init__(**data)

        self.__set_kind()
        self.__set_author()
        self.__set_recipient()
        self.__set_message()
        self.__set_isme(client)


    def __set_kind(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        command = self.command
        params = self.params

        kind: KINDS = 'event'


        if command == 'PRIVMSG':

            assert params is not None

            prefix = params[0][:1]

            kind = (
                'chanmsg'
                if prefix in '#&+!'
                else 'privmsg')


        self.kind = kind


    def __set_author(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        kind = self.kind
        prefix = self.prefix

        if kind not in MESSAGE:
            return None

        assert prefix is not None

        author = prefix.split(
            '!', maxsplit=1)

        self.author = author[0]


    def __set_recipient(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        kind = self.kind
        params = self.params

        if (kind not in MESSAGE
                or not params):
            return None

        split = params.split(' ')

        self.recipient = split[0]


    def __set_message(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        kind = self.kind
        params = self.params

        if (kind not in MESSAGE
                or not params):
            return None

        message = (
            params
            .split(':', maxsplit=1))

        self.message = message[1]


    def __set_isme(
        self,
        client: 'Client',
    ) -> None:
        """
        Update the value for the attribute from class instance.

        :param client: Class instance for connecting to service.
        """

        mynick = client.nickname
        author = self.author
        event = self.original

        operate = (
            client.params
            .operate)

        isme: bool = False


        if operate == 'service':

            prefix = f':{mynick} '

            if event.startswith(prefix):
                isme = True


        elif mynick and author:

            mine = mynick
            them = author

            isme = mine == them


        elif operate == 'normal':

            prefix = f':{mynick}!'

            if event.startswith(prefix):
                isme = NCTrue

            prefix = f':{mynick} '

            if event.startswith(prefix):
                isme = NCTrue


        self.isme = isme



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import loads
from typing import Annotated
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import BaseModel
from encommon.types import DictStrAny

from pydantic import Field

if TYPE_CHECKING:
    from .client import Client



KINDS = Literal[
    'event',
    'chanmsg',
    'privmsg']

MESSAGE = [
    'chanmsg',
    'privmsg']



class ClientEvent(BaseModel, extra='ignore'):
    """
    Contains information returned from the upstream server.
    """

    type: Annotated[
        Optional[str],
        Field(None,
              description='Type of event that occurred',
              min_length=1)]

    data: Annotated[
        Optional[DictStrAny],
        Field(None,
              description='Payload with the event data',
              min_length=1)]

    broadcast: Annotated[
        Optional[DictStrAny],
        Field(None,
              description='Payload with the event data',
              min_length=1)]

    seqno: Annotated[
        Optional[int],
        Field(None,
              description='Event number within squence',
              ge=0)]

    status: Annotated[
        Optional[str],
        Field(None,
              description='Type of event that occurred',
              min_length=1)]

    error: Annotated[
        Optional[DictStrAny],
        Field(None,
              description='Payload with the event data',
              min_length=1)]

    seqre: Annotated[
        Optional[int],
        Field(None,
              description='Reply number within squence',
              ge=0)]

    original: Annotated[
        DictStrAny,
        Field(...,
              description='Original received from server',
              min_length=1)]

    kind: Annotated[
        KINDS,
        Field('event',
              description='Dynamic field parsed from event')]

    isme: Annotated[
        bool,
        Field(False,
              description='Indicates message is from client')]

    author: Annotated[
        Optional[tuple[str, str]],
        Field(None,
              description='Dynamic field parsed from event',
              min_length=1)]

    recipient: Annotated[
        Optional[str],
        Field(None,
              description='Dynamic field parsed from event',
              min_length=1)]

    message: Annotated[
        Optional[str],
        Field(None,
              description='Dynamic field parsed from event',
              min_length=1)]


    def __init__(
        self,
        /,
        client: 'Client',
        event: DictStrAny,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        data: DictStrAny = {
            'original': event}


        type = event.get('event')
        _data = event.get('data')
        bcast = event.get('broadcast')
        seqno = event.get('seq')

        status = event.get('status')
        seqre = event.get('seq_reply')
        error = event.get('error')


        if type is not None:
            data['type'] = type

        if _data is not None:
            data['data'] = _data

        if bcast is not None:
            data['broadcast'] = bcast

        if seqno is not None:
            data['seqno'] = seqno

        if status is not None:
            data['status'] = status

        if error is not None:
            data['error'] = error

        if seqre is not None:
            data['seqre'] = seqre


        super().__init__(**data)

        self.__set_kind()
        self.__set_author()
        self.__set_recipient()
        self.__set_message()
        self.__set_isme(client)


    def __set_kind(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        type = self.type
        data = self.data

        kind: KINDS = 'event'


        if type == 'posted':

            assert data is not None

            _type = data[
                'channel_type']

            kind = (
                'privmsg'
                if _type == 'D'
                else 'chanmsg')


        self.kind = kind


    def __set_author(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        kind = self.kind
        data = self.data

        if (kind not in MESSAGE
                or not data):
            return None

        post = loads(data['post'])

        unique = post['user_id']
        name = data['sender_name']

        self.author = (
            name, unique)


    def __set_recipient(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        kind = self.kind
        data = self.data

        if (kind not in MESSAGE
                or not data):
            return None

        post = loads(data['post'])

        channel = post['channel_id']

        self.recipient = channel


    def __set_message(
        self,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        kind = self.kind
        data = self.data

        if (kind not in MESSAGE
                or not data):
            return None

        post = loads(data['post'])

        message = post['message']

        self.message = message


    def __set_isme(
        self,
        client: 'Client',
    ) -> None:
        """
        Update the value for the attribute from class instance.

        :param client: Class instance for connecting to service.
        """

        mynick = client.nickname
        author = self.author

        isme: bool = False

        if mynick and author:

            mine = mynick[1]
            them = author[1]

            isme = mine == them

        self.isme = isme



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



import asyncio
from json import loads
from typing import AsyncIterator
from typing import Iterator
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import DictStrAny

from httpx import Response

from ..utils import HTTPClient
from ..utils.http import _METHODS
from ..utils.http import _PAYLOAD

if TYPE_CHECKING:
    from .params import BridgeParams



class Bridge:
    """
    Interact with the local service API with various methods.

    :param params: Parameters used to instantiate the class.
    """

    __params: 'BridgeParams'
    __client: HTTPClient


    def __init__(
        self,
        params: 'BridgeParams',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__params = params

        client = HTTPClient(
            timeout=params.timeout,
            verify=params.ssl_verify,
            capem=params.ssl_capem)

        self.__client = client


    @property
    def params(
        self,
    ) -> 'BridgeParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def client(
        self,
    ) -> HTTPClient:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client


    def request(
        self,
        method: _METHODS,
        path: str,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
        *,
        timeout: Optional[int] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        :param timeout: Timeout waiting for the server response.
            This will override the default client instantiated.
        :returns: Response from upstream request to the server.
        """

        params = dict(params or {})
        json = dict(json or {})

        _params = self.__params
        server = _params.server
        token = _params.token

        client = self.__client

        tokey = 'hue-application-key'

        headers = {tokey: token}

        location = (
            f'https://{server}'
            f'/clip/v2/{path}')

        request = client.request_block

        return request(
            method=method,
            location=location,
            params=params,
            headers=headers,
            json=json)


    def events_block(
        self,
        timeout: Optional[int] = None,
    ) -> Iterator[DictStrAny]:
        """
        Return the response for upstream request to the server.

        :param timeout: Timeout waiting for the server response.
            This will override the default client instantiated.
        """

        _params = self.__params
        server = _params.server
        token = _params.token

        client = self.__client

        accept = 'text/event-stream'
        tokey = 'hue-application-key'

        headers = {
            'Accept': accept,
            tokey: token}

        location = (
            f'https://{server}'
            f'/eventstream/clip/v2')

        request = client.stream_block


        stream = request(
            method='get',
            location=location,
            headers=headers,
            timeout=timeout)


        for event in stream:

            if event is None:
                continue  # NOCVR

            event = event.strip()

            if event[0:5] != 'data:':
                continue

            loaded = loads(event[6:])

            for _event in loaded:

                events = _event['data']

                yield from events


    async def events_async(
        self,
        timeout: Optional[int] = None,
    ) -> AsyncIterator[DictStrAny]:
        """
        Return the response for upstream request to the server.

        :param timeout: Timeout waiting for the server response.
            This will override the default client instantiated.
        """

        _params = self.__params
        server = _params.server
        token = _params.token

        client = self.__client

        accept = 'text/event-stream'
        tokey = 'hue-application-key'

        headers = {
            'Accept': accept,
            tokey: token}

        location = (
            f'https://{server}'
            f'/eventstream/clip/v2')

        request = client.stream_async


        stream = request(
            method='get',
            location=location,
            headers=headers,
            timeout=timeout)


        async for event in stream:

            if event is None:
                continue  # NOCVR

            event = event.strip()

            if event[0:5] != 'data:':
                continue

            loaded = loads(event[6:])

            for _event in loaded:

                events = _event['data']

                for event in events:

                    assert isinstance(event, dict)

                    yield event

                    await asyncio.sleep(0)


        await asyncio.sleep(0)



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from codecs import getincrementaldecoder as getdecoder
from queue import Queue
from re import compile
from re import match as re_match
from socket import AF_INET
from socket import SOCK_STREAM
from socket import socket
from ssl import CERT_NONE
from ssl import CERT_REQUIRED
from ssl import SSLSocket
from ssl import create_default_context as default
from threading import Event
from typing import Callable
from typing import Iterator
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types.strings import NEWLINE
from encommon.types.strings import SEMPTY

from .models import ClientEvent
from ..utils import dumlog

if TYPE_CHECKING:
    from .params import ClientParams



PING = compile(
    r'^PING \:(\S+)$')

HELO = compile(
    r'^\:\S+\s001\s'
    r'(?P<crrnt>\S+)')

NICK = compile(
    r'^\:(?P<nick1>[^\!]+)'
    r'(\!\S+)?\sNICK\s\:'
    r'(?P<nick2>\S+)')



class Client:
    """
    Establish and maintain connection with the chat service.

    :param params: Parameters used to instantiate the class.
    """

    __params: 'ClientParams'
    __logger: Callable[..., None]

    __socket: Optional[socket | SSLSocket]
    __conned: Event
    __exited: Event
    __mynick: Optional[str]
    __lsnick: Optional[str]

    __mqueue: Queue[ClientEvent]
    __cancel: Event


    def __init__(
        self,
        params: 'ClientParams',
        logger: Optional[Callable[..., None]] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__params = params
        self.__logger = (
            logger or dumlog)

        self.__socket = None
        self.__conned = Event()
        self.__exited = Event()
        self.__mynick = None
        self.__lsnick = None

        self.__mqueue = Queue(
            params.queue_size)

        self.__cancel = Event()


    @property
    def params(
        self,
    ) -> 'ClientParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def connected(
        self,
    ) -> bool:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return (
            not self.__exited.is_set()
            and self.__conned.is_set())


    @property
    def nickname(
        self,
    ) -> Optional[str]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        params = self.__params

        operate = params.operate
        serverid = params.serverid

        if operate == 'service':
            return serverid

        return (
            self.__mynick
            or self.__lsnick)


    @property
    def mqueue(
        self,
    ) -> Queue[ClientEvent]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__mqueue


    @property
    def canceled(
        self,
    ) -> bool:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return (
            self.__cancel.is_set()
            or self.__exited.is_set())


    def operate(
        self,
    ) -> None:
        """
        Operate the client and populate queue with the messages.
        """

        logger = self.__logger

        try:

            logger(item='initial')

            self.__socket = None
            self.__conned.clear()
            self.__exited.clear()
            self.__mynick = None

            self.__cancel.clear()

            logger(item='operate')

            self.__operate()

        finally:

            self.__socket = None
            self.__conned.clear()
            self.__exited.clear()
            self.__mynick = None

            self.__cancel.clear()

            logger(item='finish')


    def __operate(
        self,
    ) -> None:
        """
        Operate the client and populate queue with the messages.
        """

        logger = self.__logger


        self.__connect()

        socket = self.__socket

        assert socket is not None


        while not self.canceled:

            receive = (
                self.socket_recv())

            for event in receive:
                self.__event(event)


        logger(item='close')

        socket.close()


        if self.__exited.is_set():
            raise ConnectionError


    def __event(
        self,
        event: str,
    ) -> None:
        """
        Operate the client and populate queue with the messages.

        :param event: Raw event received from the network peer.
        """

        params = self.__params
        logger = self.__logger
        mqueue = self.__mqueue
        crrnt = self.__mynick

        model = ClientEvent

        operate = params.operate


        if operate == 'normal':


            match = re_match(
                HELO, event)

            if match is not None:

                crrnt = (
                    match.group('crrnt'))

                logger(item='helo')

                self.__mynick = crrnt
                self.__lsnick = crrnt


            match = re_match(
                NICK, event)

            if match is not None:

                nick1 = (
                    match.group('nick1'))

                nick2 = (
                    match.group('nick2'))

                if nick1 == crrnt:
                    self.__mynick = nick2
                    self.__lsnick = nick2


        match = re_match(
            PING, event)

        if match is not None:

            ping = match.group(1)
            pong = f'PONG {ping}'

            logger(item='ping')

            self.socket_send(pong)

            return None


        object = model(
            self, event)

        mqueue.put(object)


    def stop(
        self,
    ) -> None:
        """
        Gracefully close the connection with the server socket.
        """

        logger = self.__logger

        logger(item='stop')

        if self.connected:

            self.socket_send(
                'QUIT :Adios')

        self.__cancel.set()


    def __wrapper(
        self,
        socket: socket,
    ) -> SSLSocket:
        """
        Construct the object from wrapping SSL context settings.

        :param socket: Socket instance that will be SSL wrapped.
        """

        params = self.__params

        server = params.server
        verify = params.ssl_verify

        context = default()


        setattr(
            context,
            'check_hostname',
            verify)


        _verify = (
            CERT_REQUIRED
            if verify is True
            else CERT_NONE)

        setattr(
            context,
            'verify_mode',
            _verify)


        wrapper = context.wrap_socket

        return wrapper(
            socket,
            server_hostname=server)


    def __connect(
        self,
    ) -> None:
        """
        Establish the connection with the upstream using socket.
        """

        params = self.__params
        logger = self.__logger

        server = params.server
        port = params.port
        operate = params.operate
        nick = params.nickname
        user = params.username
        real = params.realname
        passwd = params.password
        sname = params.servername
        serverid = params.serverid
        senable = params.ssl_enable

        address = (server, port)


        handle = socket(
            AF_INET, SOCK_STREAM)

        handle.settimeout(1)


        if senable is True:
            wrapper = self.__wrapper
            handle = wrapper(handle)


        logger(item='connect')

        handle.connect(address)


        self.__socket = handle


        if operate == 'normal':

            self.socket_send(
                f'NICK {nick}')

            self.socket_send(
                f'USER {user} . '
                f'{nick} :{real}')


        if passwd is not None:

            self.socket_send(
                f'PASS {passwd}')


        if operate == 'service':

            self.socket_send(
                'PROTOCTL'
                f' EAUTH={sname}'
                f' SID={serverid}')

            self.socket_send(
                'PROTOCTL'
                ' NOQUIT'
                ' NICKv2'
                ' SJOIN'
                ' SJ3'
                ' NICKIP'
                ' TKLEXT'
                ' TKLEXT2'
                ' NEXTBANS'
                ' CLK'
                ' ESVID'
                ' MLOCK'
                ' MTAGS'
                ' EXTSWHOIS')

            self.socket_send(
                f'SERVER {sname}'
                f' 1 :{real}')


        self.__conned.set()
        self.__exited.clear()

        self.__mynick = None


    def socket_send(
        self,
        send: str,
    ) -> None:
        """
        Transmit provided content through the socket connection.

        :param send: Content which will be sent through socket.
        """

        logger = self.__logger
        socket = self.__socket

        assert socket is not None

        logger(
            item='transmit',
            value=send)

        transmit = (
            f'{send}\r\n'
            .encode('utf-8'))

        socket.send(transmit)


    def socket_recv(
        self,
    ) -> Iterator[str]:
        """
        Return the content received from the socket connection.

        .. note::
           Method could be further optimized using an internal
           buffer within the class after receiving more bytes.

        :returns: Content received from the socket connection.
        """

        logger = self.__logger
        exited = self.__exited
        socket = self.__socket

        assert socket is not None

        buffer = bytearray()

        decoder = getdecoder('utf-8')()

        bempty = (
            SEMPTY
            .encode('utf-8'))

        bewline = (
            NEWLINE
            .encode('utf-8'))


        def _receive() -> bytes:

            return socket.recv(4096)


        while True:

            try:

                recv = _receive()

                if recv == bempty:
                    exited.set()
                    break

                buffer.extend(recv)

            except TimeoutError:
                break


            while True:

                index = (
                    buffer.find(bewline))

                if index == -1:
                    break

                line = buffer[:index + 1]

                del buffer[:index + 1]

                event = (
                    decoder.decode(
                        line, False)
                    .strip('\r\n'))

                if event[:5] == 'ERROR':
                    exited.set()

                if len(event) >= 1:

                    logger(
                        item='receive',
                        value=event)

                    yield event
